/******************************************************************

 Copyright 2001 by Object Craft P/L, Melbourne, Australia.

 LICENCE - see LICENCE file distributed with this software for details.

******************************************************************/

#include "mssqldb.h"

static PyObject *ErrorObject;

static PyObject *debug_file = NULL;

static char mssqldb_DataBuf__doc__[] =
"DataBuf(type [, len]) -> databuf";

static PyObject *mssqldb_DataBuf(PyObject *self, PyObject *args)
{
    int vartype, varlen = 0;

    if (!PyArg_ParseTuple(args, "i|i", &vartype, &varlen))
	return NULL;

    return DataBuf__init__(vartype, varlen);
}

static char mssqldb_dbconvert__doc__[] =
"dbconvert(obj, databuf) -> status";

static PyObject *mssqldb_dbconvert(PyObject *self, PyObject *args)
{
    PyObject *from;
    DataBufObj *to;

    if (!PyArg_ParseTuple(args, "OO!", &from, &DataBufType, &to))
	return NULL;
    return DataBuf_convert(NULL, from, to);
}

static PyObject *err_handler_cb;

static int err_handler(DBPROCESS *db_dbproc,
		       int severity, int dberr, int oserr,
		       const char *dberrstr, const char *oserrstr)
{
    PyObject *dbproc, *args, *result;

    if (db_dbproc == NULL)
	dbproc = Py_None;
    else {
	dbproc = dbproc_find_object(db_dbproc);
	if (dbproc == NULL)
	    return INT_CANCEL;
    }
    if (oserr == DBNOERR)
	oserrstr = "";
    args = Py_BuildValue("(Oiiiss)",
			 dbproc, severity, dberr, oserr, dberrstr, oserrstr);
    if (args == NULL)
	return INT_CANCEL;
    result = PyEval_CallObject(err_handler_cb, args);
    Py_DECREF(args);
    if (result != NULL) {
	RETCODE status = INT_CANCEL;

	if (PyInt_Check(result))
	    status = PyInt_AsLong(result);
	Py_DECREF(result);
	return status;
    }

    return INT_CANCEL;
}

static PyObject *msg_handler_cb;

static int msg_handler(DBPROCESS *db_dbproc,
		       DBINT msgno, int msgstate, int severity,
		       const char *msgtext, const char *srvname, const char *procname,
		       DBUSMALLINT line)
{
    PyObject *dbproc, *args, *result;

    if (db_dbproc == NULL)
	dbproc = Py_None;
    else {
	dbproc = dbproc_find_object(db_dbproc);
	if (dbproc == NULL)
	    dbproc = Py_None;
    }
    args = Py_BuildValue("(Oiiisssi)",
			 dbproc, msgno, msgstate, severity,
			 msgtext, srvname, procname, line);
    if (args == NULL)
	return 0;
    result = PyEval_CallObject(msg_handler_cb, args);
    Py_DECREF(args);
    if (result != NULL) {
	RETCODE status = 0;

	if (PyInt_Check(result))
	    status = PyInt_AsLong(result);
	Py_DECREF(result);
	return status;
    }

    return 0;
}

static char mssqldb_dberrhandle__doc__[] =
"dberrhandle(callable) -> callable";

static PyObject *mssqldb_dberrhandle(PyObject *self, PyObject *args)
{
    static int handler_set;
    PyObject *func;
    PyObject *old_func;

    if (!PyArg_ParseTuple(args, "O", &func))
	return NULL;

    old_func = err_handler_cb;
    if (func == Py_None)
	err_handler_cb = NULL;
    else {
	if (!PyCallable_Check(func)) {
	    PyErr_SetString(PyExc_TypeError, "parameter must be callable");
	    return NULL;
	}
	err_handler_cb = func;
	Py_XINCREF(err_handler_cb);
    }

    if (!handler_set) {
	handler_set = 1;
	PyErr_Clear();
#ifdef _WIN32
	dberrhandle(err_handler);
#else
	dberrhandle((EHANDLEFUNC)err_handler);
#endif
	debug_msg("dberrhandle()\n");
	if (PyErr_Occurred())
	    return NULL;
    }

    if (old_func != NULL)
	return old_func;
    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbexit__doc__[] =
"dbexit() -> None";

static PyObject *mssqldb_dbexit(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    dbexit();
    debug_msg("dbexit()\n");
    if (PyErr_Occurred())
	return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbinit__doc__[] =
"dbinit() -> bool";

#ifdef _WIN32
static PyObject *mssqldb_dbinit(PyObject *self, PyObject *args)
{
    const char *version;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    version = dbinit();
    debug_msg("dbinit() -> '%s'\n", version);
    if (PyErr_Occurred())
	return NULL;

    if (version == NULL) {
	Py_INCREF(Py_None);
	return Py_None;
    }

    return PyString_FromString(version);
}
#else
static PyObject *mssqldb_dbinit(PyObject *self, PyObject *args)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    status = dbinit();
    debug_msg("dbinit() -> %s\n", value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
#endif

static char mssqldb_dblogin__doc__[] =
"dblogin(debug = 0) -> LOGINREC";

static PyObject *mssqldb_dblogin(PyObject *self, PyObject *args)
{
    int debug = (debug_file != NULL && debug_file != Py_None);

    if (!PyArg_ParseTuple(args, "|i", &debug))
	return NULL;

    return LOGINREC__init__(debug);
}

static char mssqldb_dbmsghandle__doc__[] =
"dbmsghandle(callable) -> callable";

static PyObject *mssqldb_dbmsghandle(PyObject *self, PyObject *args)
{
    static int handler_set;
    PyObject *func;
    PyObject *old_func;

    if (!PyArg_ParseTuple(args, "O", &func))
	return NULL;

    old_func = msg_handler_cb;
    if (func == Py_None)
	msg_handler_cb = NULL;
    else {
	if (!PyCallable_Check(func)) {
	    PyErr_SetString(PyExc_TypeError, "parameter must be callable");
	    return NULL;
	}
	msg_handler_cb = func;
	Py_XINCREF(msg_handler_cb);
    }

    if (!handler_set) {
	handler_set = 1;
	PyErr_Clear();
#ifdef _WIN32
	dbmsghandle(msg_handler);
#else
	dbmsghandle((MHANDLEFUNC)msg_handler);
#endif
	debug_msg("dbmsghandle()\n");
	if (PyErr_Occurred())
	    return NULL;
    }

    if (old_func != NULL)
	return old_func;
    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbprtype__doc__[] = 
"dbprtype(token) -> name";

static PyObject *mssqldb_dbprtype(DBPROCObj *self, PyObject *args)
{
    int token;
    const char *str;

    if (!PyArg_ParseTuple(args, "i", &token))
	return NULL;

    str = dbprtype(token);
    debug_msg("dbprtype(%d) -> %s\n", token, str);

    if (str == NULL) {
	Py_INCREF(Py_None);
	return Py_None;
    }
    return PyString_FromString(str);
}

void debug_msg(char *fmt, ...)
{
    char buff[10240];
    va_list ap;
    PyObject *res;

    if (debug_file == Py_None)
	return;

    va_start(ap, fmt);
#ifdef _WIN32
    _vsnprintf(buff, sizeof(buff), fmt, ap);
#else
    vsnprintf(buff, sizeof(buff), fmt, ap);
#endif
    va_end(ap);

    res = PyObject_CallMethod(debug_file, "write", "s", buff);
    Py_XDECREF(res);
    PyErr_Clear();
    res = PyObject_CallMethod(debug_file, "flush", "");
    Py_XDECREF(res);
    PyErr_Clear();
}

static char mssqldb_set_debug__doc__[] = 
"set_debug(file) -> old_file";

static PyObject *mssqldb_set_debug(DBPROCObj *self, PyObject *args)
{
    PyObject *obj, *res;

    if (!PyArg_ParseTuple(args, "O", &obj))
	return NULL;

    if (obj != Py_None) {
	res = PyObject_CallMethod(obj, "write", "s", "");
	Py_XDECREF(res);
	if (res == NULL)
	    return NULL;
    }

    res = debug_file;
    debug_file = obj;
    Py_INCREF(debug_file);
    return res;
}

static struct PyMethodDef mssqldb_methods[] = {
    { "DataBuf", (PyCFunction)mssqldb_DataBuf, METH_VARARGS, mssqldb_DataBuf__doc__ },
    { "datetime", (PyCFunction)DateTimeType_new, METH_VARARGS, datetime_new__doc__ },
    { "dbconvert", (PyCFunction)mssqldb_dbconvert, METH_VARARGS, mssqldb_dbconvert__doc__ },
    { "dberrhandle", (PyCFunction)mssqldb_dberrhandle, METH_VARARGS, mssqldb_dberrhandle__doc__ },
    { "dbexit", (PyCFunction)mssqldb_dbexit, METH_VARARGS, mssqldb_dbexit__doc__ },
    { "dbinit", (PyCFunction)mssqldb_dbinit, METH_VARARGS, mssqldb_dbinit__doc__ },
    { "dblogin", (PyCFunction)mssqldb_dblogin, METH_VARARGS, mssqldb_dblogin__doc__ },
    { "dbmsghandle", (PyCFunction)mssqldb_dbmsghandle, METH_VARARGS, mssqldb_dbmsghandle__doc__ },
    { "dbprtype", (PyCFunction)mssqldb_dbprtype, METH_VARARGS, mssqldb_dbprtype__doc__ },
    { "money", (PyCFunction)MoneyType_new, METH_VARARGS, money_new__doc__ },
    { "numeric", (PyCFunction)NumericType_new, METH_VARARGS, numeric_new__doc__ },
    { "pickle_datetime", (PyCFunction)pickle_datetime, METH_VARARGS, pickle_datetime__doc__ },
    { "pickle_money", (PyCFunction)pickle_money, METH_VARARGS, pickle_money__doc__ },
    { "pickle_numeric", (PyCFunction)pickle_numeric, METH_VARARGS, pickle_numeric__doc__ },
    { "set_debug", (PyCFunction)mssqldb_set_debug, METH_VARARGS, mssqldb_set_debug__doc__ },
    { NULL, (PyCFunction)NULL, 0, NULL }
};

#define DBVAL(t, v) { t, #v, v }

typedef struct {
    int type;
    char *name;
    int value;
} value_desc;

static value_desc mssql_args[] = {
    DBVAL(TYPE_RETCODE, FAIL),
    DBVAL(TYPE_RETCODE, SUCCEED),
    DBVAL(TYPE_RETCODE, NO_MORE_RESULTS),

    DBVAL(TYPE_ANSWER, INT_EXIT),
#ifdef INT_TIMEOUT
    DBVAL(TYPE_ANSWER, INT_CONTINUE),
#endif
    DBVAL(TYPE_ANSWER, INT_CANCEL),
#ifdef INT_TIMEOUT
    DBVAL(TYPE_ANSWER, INT_TIMEOUT),
#endif

    DBVAL(TYPE_ROWRES, MORE_ROWS),
    DBVAL(TYPE_ROWRES, NO_MORE_ROWS),
    DBVAL(TYPE_ROWRES, REG_ROW),
    DBVAL(TYPE_ROWRES, BUF_FULL),

    DBVAL(TYPE_VALUE, SQLMAXNAME),

#ifdef SQLVOID
    DBVAL(TYPE_SERVTYPE, SQLVOID),
#endif
    DBVAL(TYPE_SERVTYPE, SQLTEXT),
    DBVAL(TYPE_SERVTYPE, SQLVARBINARY),
    DBVAL(TYPE_SERVTYPE, SQLINTN),
    DBVAL(TYPE_SERVTYPE, SQLVARCHAR),
    DBVAL(TYPE_SERVTYPE, SQLBINARY),
    DBVAL(TYPE_SERVTYPE, SQLIMAGE),
    DBVAL(TYPE_SERVTYPE, SQLCHAR),
    DBVAL(TYPE_SERVTYPE, SQLINT1),
    DBVAL(TYPE_SERVTYPE, SQLBIT),
    DBVAL(TYPE_SERVTYPE, SQLINT2),
    DBVAL(TYPE_SERVTYPE, SQLINT4),
    DBVAL(TYPE_SERVTYPE, SQLMONEY),
    DBVAL(TYPE_SERVTYPE, SQLDATETIME),
    DBVAL(TYPE_SERVTYPE, SQLFLT8),
    DBVAL(TYPE_SERVTYPE, SQLFLTN),
    DBVAL(TYPE_SERVTYPE, SQLMONEYN),
    DBVAL(TYPE_SERVTYPE, SQLDATETIMN),
    DBVAL(TYPE_SERVTYPE, SQLFLT4),
    DBVAL(TYPE_SERVTYPE, SQLMONEY4),
    DBVAL(TYPE_SERVTYPE, SQLDATETIM4),
    DBVAL(TYPE_SERVTYPE, SQLDECIMAL),
    DBVAL(TYPE_SERVTYPE, SQLNUMERIC),

#ifdef DBRPCRECOMPILE
    DBVAL(TYPE_RPCOPT, DBRPCRECOMPILE),
#endif
#ifdef DBNOTIFYALL
    DBVAL(TYPE_RPCOPT, DBNOTIFYALL),
#endif
#ifdef DBNOTIFYNEXT
    DBVAL(TYPE_RPCOPT, DBNOTIFYNEXT),
#endif

#ifdef DBRPCNORETURN
    DBVAL(TYPE_RPCPARAM, DBRPCNORETURN),
#endif
    DBVAL(TYPE_RPCPARAM, DBRPCRETURN),
#ifdef DB__NODEFAULT
    DBVAL(TYPE_RPCPARAM, DB__NODEFAULT),
#endif

    DBVAL(TYPE_BIND, BINARYBIND),
    DBVAL(TYPE_BIND, STRINGBIND),
    DBVAL(TYPE_BIND, BITBIND),
    DBVAL(TYPE_BIND, TINYBIND),
    DBVAL(TYPE_BIND, SMALLBIND),
    DBVAL(TYPE_BIND, INTBIND),
    DBVAL(TYPE_BIND, MONEYBIND),
    DBVAL(TYPE_BIND, FLT4BIND),
    DBVAL(TYPE_BIND, FLT8BIND),
    DBVAL(TYPE_BIND, DATETIMEBIND),
    DBVAL(TYPE_BIND, NUMERICBIND),
    DBVAL(TYPE_BIND, DECIMALBIND),

    DBVAL(TYPE_OPTION, DBPARSEONLY),
#ifdef DBESTIMATE
    DBVAL(TYPE_OPTION, DBESTIMATE),       
#endif
    DBVAL(TYPE_OPTION, DBSHOWPLAN),
    DBVAL(TYPE_OPTION, DBNOEXEC),
    DBVAL(TYPE_OPTION, DBARITHIGNORE),
    DBVAL(TYPE_OPTION, DBNOCOUNT),
    DBVAL(TYPE_OPTION, DBARITHABORT),
    DBVAL(TYPE_OPTION, DBTEXTLIMIT),
#ifdef DBBROWSE
    DBVAL(TYPE_OPTION, DBBROWSE),
#endif
    DBVAL(TYPE_OPTION, DBOFFSET),
    DBVAL(TYPE_OPTION, DBSTAT),
#ifdef DBERRLVL
    DBVAL(TYPE_OPTION, DBERRLVL),
#endif
#ifdef DBCONFIRM
    DBVAL(TYPE_OPTION, DBCONFIRM),
#endif
    DBVAL(TYPE_OPTION, DBSTORPROCID),
    DBVAL(TYPE_OPTION, DBBUFFER),
    DBVAL(TYPE_OPTION, DBNOAUTOFREE),
    DBVAL(TYPE_OPTION, DBROWCOUNT),
    DBVAL(TYPE_OPTION, DBTEXTSIZE),
#ifdef DBNATLANG
    DBVAL(TYPE_OPTION, DBNATLANG),
#endif
#ifdef DBDATEFORMAT
    DBVAL(TYPE_OPTION, DBDATEFORMAT),
#endif
#ifdef DBPRPAD
    DBVAL(TYPE_OPTION, DBPRPAD),
#endif
#ifdef DBPRCOLSEP
    DBVAL(TYPE_OPTION, DBPRCOLSEP),
#endif
#ifdef DBPRLINELEN
    DBVAL(TYPE_OPTION, DBPRLINELEN),
#endif
#ifdef DBPRLINESEP
    DBVAL(TYPE_OPTION, DBPRLINESEP),
#endif
#ifdef DBLFCONVERT
    DBVAL(TYPE_OPTION, DBLFCONVERT),
#endif
#ifdef DBDATEFIRST
    DBVAL(TYPE_OPTION, DBDATEFIRST),
#endif
#ifdef DBCHAINXACTS
    DBVAL(TYPE_OPTION, DBCHAINXACTS),
#endif
#ifdef DBFIPSFLAG
    DBVAL(TYPE_OPTION, DBFIPSFLAG),
#endif
#ifdef DBISOLATION
    DBVAL(TYPE_OPTION, DBISOLATION),
#endif
#ifdef DBAUTH
    DBVAL(TYPE_OPTION, DBAUTH),
#endif
#ifdef DBIDENTITY
    DBVAL(TYPE_OPTION, DBIDENTITY),
#endif
#ifdef DBNOIDCOL
    DBVAL(TYPE_OPTION, DBNOIDCOL),
#endif
#ifdef DBDATESHORT
    DBVAL(TYPE_OPTION, DBDATESHORT),
#endif

    {0,0,0}
};

char *value_str(int type, int value)
{
    value_desc *desc;
    char *name = NULL;
    static char num_str[16];

    for (desc = mssql_args; desc->name != NULL; desc++)
	if (desc->value == value) {
	    if (desc->type == type)
		return desc->name;
	    name = desc->name;
	}
    if (name != NULL)
	return name;
    sprintf(num_str, "%d", value);
    return num_str;
}

static int dict_add_int(PyObject *dict, char *key, int value)
{
    int err;
    PyObject *obj = PyInt_FromLong(value);
    if (obj == NULL)
	return 0;

    err = PyDict_SetItemString(dict, key, obj);
    Py_DECREF(obj);
    return !err;
}

static int dict_add_type(PyObject *dict, PyTypeObject *type)
{
    return PyDict_SetItemString(dict, type->tp_name, (PyObject*)type);
}

static char mssqldb_module_documentation[] = 
"";

void initmssqldb(void)
{
    PyObject *m, *d, *rev = NULL;
    value_desc *desc;

    LOGINRECType.ob_type = &PyType_Type;
    DBPROCType.ob_type = &PyType_Type;
    DataBufType.ob_type = &PyType_Type;
    NumericType.ob_type = &PyType_Type;
    DateTimeType.ob_type = &PyType_Type;
    MoneyType.ob_type = &PyType_Type;

    m = Py_InitModule4("mssqldb", mssqldb_methods,
		       mssqldb_module_documentation,
		       (PyObject*)NULL, PYTHON_API_VERSION);

    d = PyModule_GetDict(m);
    /* Add constants */
    for (desc = mssql_args; desc->name != NULL; desc++)
	if (!dict_add_int(d, desc->name, desc->value))
	    break;

    /* set debug file */
    debug_file = Py_None;
    Py_INCREF(debug_file);

    /* Add type objects */
    if (dict_add_type(d, &LOGINRECType)
	|| dict_add_type(d, &DBPROCType)
	|| dict_add_type(d, &DataBufType)
	|| dict_add_type(d, &DateTimeType)
	|| dict_add_type(d, &NumericType))
	;

    /* Register the numeric pickler */
    copy_reg_numeric(d);
    copy_reg_datetime(d);
    copy_reg_money(d);

    ErrorObject = PyString_FromString("mssqldb.error");
    PyDict_SetItemString(d, "error", ErrorObject);

    rev = PyString_FromString("0.09");
    PyDict_SetItemString(d, "__version__", rev);
    Py_DECREF(rev);

    if (PyErr_Occurred())
	Py_FatalError("can't initialize module mssqldb");
}

